'' =================================================================================================
''
''   File....... jm_pwm8.spin
''   Purpose.... LED modulation using fixed-frequency pwm (up to ~360Hz @ 80MHz sys clock)
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (C) 2011-2017 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started.... 
''   Updated.... 04 MAR 2017
''
'' =================================================================================================


con { fixed io pins }
                                                                 
  RX1  = 31                                                     ' programming / terminal
  TX1  = 30                                                       
                                                                 
  SDA1 = 29                                                     ' eeprom / i2c
  SCL1 = 28                                                       
                                                                 
                                                                 
con                                                              
                                                                 
  MIN_TIX = 416 { 750Hz @ 80MHz sys clock }                     ' cnt ticks per pwm loop
                                                                 

con                                                              
                                                                 
  CHANNELS = 8                                                   
  LAST_CH  = CHANNELS - 1                                        
                                                                 
                                                                 
var                                                              
                                                                 
  long  cog                                                      

  long  setup
  long  brpntr                                                   
                                                                 
  byte  brightness[CHANNELS]                                    ' brightness levels
                                                                 
                                                                 
pub null                                                         

  ' This is not a top-level object


pub start(count, base)

'' Start PWM8 pwm driver; uses one cog
'' -- count is number of pwm outputs (1 to CHANNELS)
'' -- base is the first pin to use

  return startx(count, base, 360)                               ' pwm @ 360Hz

  
pub startx(count, base, freq)

'' Start PWM8 pwm driver; uses one cog
'' -- n is number of pwm outputs (1 to 8)
'' -- p is the first pin to use
'' -- freq is the modulation frequency

  stop                                                          ' stop if running

  setup.byte[0] := 1 #> count <#8
  setup.byte[1] := base
  setup.word[1] := ((clkfreq / freq) >> 8) #> MIN_TIX   
  brpntr        := @brightness                                      
                                                             
  cog := cognew(@pwm8, @setup) + 1                           ' start the pwm cog
                                                                 
  return cog                                                     
                                                                 
                                                                 
pub stop                                                         
                                                                 
'' Stops PWM8 driver; frees a cog                                
                                                                 
  set_all(0)                                                     
                                                                 
  if (cog)                                                       
    cogstop(cog - 1)                                             
    cog := 0                                                     
                                                                 
                                                                 
pub running                                                      
                                                                 
'' Reports true if PWM cog is loaded                             
                                                                 
  return (cog > 0)                                               
                                                                 

pub set(ch, level)

'' Sets channel to specified level

  if ((ch => 0) and (ch =< LAST_CH))
    brightness[ch] := 0 #> level <# 255

  return brightness[ch]
    

pub set_all(level)

'' Set all channels to same level

  level := 0 #> level <# 255                                    ' limit to byte value
  bytefill(@brightness[0], level, CHANNELS)

  return level


pub high(ch)

'' Sets channel to on

  if ((ch => 0) and (ch =< LAST_CH))
    brightness[ch] := 255


pub low(ch)

'' Sets channel to off

  if ((ch => 0) and (ch =< LAST_CH))
    brightness[ch] := 0


pub toggle(ch)

'' Inverts channel level

  if ((ch => 0) and (ch =< LAST_CH))
    if (brightness[ch] & $80)
      brightness[ch] := 0
    else
      brightness[ch] := 255

  
pub inc(ch)

'' Increment channel brightness

  if ((ch => 0) and (ch =< LAST_CH))
    if (brightness[ch] < 255)
      ++brightness[ch]

  return brightness[ch]


pub inc_all | ch

'' Increment all channels

  repeat ch from 0 to LAST_CH
    if (brightness[ch] < 255)
      ++brightness[ch]


pub dec(ch)

'' Decrement channel brightness

  if ((ch => 0) and (ch =< LAST_CH))
    if (brightness[ch] > 0)
      --brightness[ch]

  return brightness[ch]


pub dec_all | ch

'' Decrement all channels

  repeat ch from 0 to LAST_CH
    if (brightness[ch] > 0)
      --brightness[ch]


pub digital(value, mask) | ch

'' Sets enabled pins (in mask) to corresponding bit in value

  repeat ch from 0 to LAST_CH                                   ' loop through outputs
    if (mask & (1 << ch))                                       ' if channel enabled
      if (value & (1 << ch))                                    ' if on
        brightness[ch] := 255                                   '  set to full 
      else                                                       
        brightness[ch] := 0                                     '  set to off
                                                                 
                                                                 
pub dcd(bit)                                                     
                                                                 
'' Sets output to decoded power-of-2 value                       
                                                                 
  if (bit < 0)                                                  ' if negative
    set_all(0)                                                  '  all off
  elseif (bit =< LAST_CH)                                       ' if valid
    digital(1 << bit, %1111_1111)                               '  convert to bit output
                                                                 

pub gamma8(level)

  return GammaTable[0 #> level <# 255]


pub read(ch)

'' Returns channel brightness

  if ((ch => 0) and (ch =< LAST_CH))
    return brightness[ch]
  else
    return -1


pub address

'' Returns hub address of brightness array

  return @brightness


dat { gamma table }

' Liberated from an Adafruit WS2812 demo

  GammaTable            byte      0,   0,   0,   0,   0,   0,   0,   0 
                        byte      0,   0,   0,   0,   0,   0,   0,   0
                        byte      0,   0,   0,   0,   0,   0,   0,   0
                        byte      1,   1,   1,   1,   1,   1,   1,   1
                        byte      1,   1,   1,   1,   2,   2,   2,   2
                        byte      2,   2,   2,   2,   3,   3,   3,   3
                        byte      3,   3,   4,   4,   4,   4,   5,   5
                        byte      5,   5,   5,   6,   6,   6,   6,   7
                        byte      7,   7,   8,   8,   8,   9,   9,   9
                        byte     10,  10,  10,  11,  11,  11,  12,  12
                        byte     13,  13,  13,  14,  14,  15,  15,  16
                        byte     16,  17,  17,  18,  18,  19,  19,  20
                        byte     20,  21,  21,  22,  22,  23,  24,  24
                        byte     25,  25,  26,  27,  27,  28,  29,  29
                        byte     30,  31,  31,  32,  33,  34,  34,  35
                        byte     36,  37,  38,  38,  39,  40,  41,  42
                        byte     42,  43,  44,  45,  46,  47,  48,  49
                        byte     50,  51,  52,  53,  54,  55,  56,  57
                        byte     58,  59,  60,  61,  62,  63,  64,  65
                        byte     66,  68,  69,  70,  71,  72,  73,  75
                        byte     76,  77,  78,  80,  81,  82,  84,  85
                        byte     86,  88,  89,  90,  92,  93,  94,  96
                        byte     97,  99, 100, 102, 103, 105, 106, 108
                        byte    109, 111, 112, 114, 115, 117, 119, 120
                        byte    122, 124, 125, 127, 129, 130, 132, 134
                        byte    136, 137, 139, 141, 143, 145, 146, 148
                        byte    150, 152, 154, 156, 158, 160, 162, 164
                        byte    166, 168, 170, 172, 174, 176, 178, 180
                        byte    182, 184, 186, 188, 191, 193, 195, 197
                        byte    199, 202, 204, 206, 209, 211, 213, 215
                        byte    218, 220, 223, 225, 227, 230, 232, 235
                        byte    237, 240, 242, 245, 247, 250, 252, 255


dat

                        org     0

pwm8                    mov     t1, par                         ' point to setup
                        rdlong  t2, t1                          ' read

                        mov     chcount, t2                     ' extract count
                        and     chcount, #$0F

                        mov     ch0pin, t2                      ' extract lsb pin
                        shr     ch0pin, #8
                        and     ch0pin, #$1F

                        mov     bittime, t2                     ' extract bit timing
                        shr     bittime, #16
                                                                 
                        add     t1, #4                           
                        rdlong  hubch0, t1                      ' read hub addr of levels array
                                                                 
                        mov     t1, #%1111_1111                 ' create mask for pins
                        mov     t2, #8                           
                        sub     t2, chcount                      
                        shr     t1, t2                          ' removed unused
                        shl     t1, ch0pin                      ' align lsbs
                        mov     dira, t1                        ' set outputs     
                                                                 
                        mov     bridx, #0                       ' reset brightness index
                                                                 
                        mov     timer, bittime                  ' start bit timer
                        add     timer, cnt                      ' sync with system timer

                        ' unrolled for speed
                        
ch0                     mov     chmask, #1                      ' reset channel output mask
                        shl     chmask, ch0pin                     
                        mov     chpntr, hubch0                  ' reset hub pointer for levels 
                        rdbyte  t1, chpntr                      ' read channel level  
                        cmp     t1, #0                  wz      ' off?
        if_z            andn    outa, chmask                    ' yes, kill output
        if_z            jmp     #ch0exit                        '  and exit
                        cmp     bridx, t1               wc, wz  ' check level
        if_be           or      outa, chmask                    '  on if not done
        if_a            andn    outa, chmask                    '  else kill output

ch0exit                 cmp     chcount, #1             wz
        if_e            jmp     #looppad


ch1                     shl     chmask, #1               
                        add     chpntr, #1              
                        rdbyte  t1, chpntr                     
                        cmp     t1, #0                  wz     
        if_z            andn    outa, chmask                   
        if_z            jmp     #ch1exit                       
                        cmp     bridx, t1               wc, wz 
        if_be           or      outa, chmask                   
        if_a            andn    outa, chmask                   
                                                               
ch1exit                 cmp     chcount, #2             wz     
        if_e            jmp     #looppad                       


ch2                     shl     chmask, #1               
                        add     chpntr, #1              
                        rdbyte  t1, chpntr                     
                        cmp     t1, #0                  wz     
        if_z            andn    outa, chmask                   
        if_z            jmp     #ch2exit                       
                        cmp     bridx, t1               wc, wz 
        if_be           or      outa, chmask                   
        if_a            andn    outa, chmask                   
                                                               
ch2exit                 cmp     chcount, #3             wz     
        if_e            jmp     #looppad 


ch3                     shl     chmask, #1               
                        add     chpntr, #1              
                        rdbyte  t1, chpntr                     
                        cmp     t1, #0                  wz     
        if_z            andn    outa, chmask                   
        if_z            jmp     #ch3exit                       
                        cmp     bridx, t1               wc, wz 
        if_be           or      outa, chmask                   
        if_a            andn    outa, chmask                   
                                                               
ch3exit                 cmp     chcount, #4             wz     
        if_e            jmp     #looppad


ch4                     shl     chmask, #1               
                        add     chpntr, #1              
                        rdbyte  t1, chpntr                     
                        cmp     t1, #0                  wz     
        if_z            andn    outa, chmask                   
        if_z            jmp     #ch4exit                       
                        cmp     bridx, t1               wc, wz 
        if_be           or      outa, chmask                   
        if_a            andn    outa, chmask                   
                                                               
ch4exit                 cmp     chcount, #5             wz     
        if_e            jmp     #looppad


ch5                     shl     chmask, #1               
                        add     chpntr, #1              
                        rdbyte  t1, chpntr                     
                        cmp     t1, #0                  wz     
        if_z            andn    outa, chmask                   
        if_z            jmp     #ch5exit                       
                        cmp     bridx, t1               wc, wz 
        if_be           or      outa, chmask                   
        if_a            andn    outa, chmask                   
                                                               
ch5exit                 cmp     chcount, #6             wz     
        if_e            jmp     #looppad


ch6                     shl     chmask, #1               
                        add     chpntr, #1              
                        rdbyte  t1, chpntr                     
                        cmp     t1, #0                  wz     
        if_z            andn    outa, chmask                   
        if_z            jmp     #ch6exit                       
                        cmp     bridx, t1               wc, wz 
        if_be           or      outa, chmask                   
        if_a            andn    outa, chmask                   
                                                               
ch6exit                 cmp     chcount, #7             wz     
        if_e            jmp     #looppad


ch7                     shl     chmask, #1               
                        add     chpntr, #1              
                        rdbyte  t1, chpntr                     
                        cmp     t1, #0                  wz     
        if_z            andn    outa, chmask                   
        if_z            jmp     #ch7exit                       
                        cmp     bridx, t1               wc, wz 
        if_be           or      outa, chmask                   
        if_a            andn    outa, chmask                   
                                                               
ch7exit                 cmp     chcount, #8             wz     
        if_e            jmp     #looppad

                                                                 
looppad                 add     bridx, #1                       ' update brightness index
                        and     bridx, #$FF                     ' rollover
                                                                 
                        waitcnt timer, bittime                     
                        jmp     #ch0                       

' --------------------------------------------------------------------------------------------------                        

chcount                 res     1                               ' channel count
ch0pin                  res     1                               ' channel 0 pin #
bittime                 res     1                               ' timing for each bit
hubch0                  res     1                               ' hub address of brightness array                     
                                                                 
bridx                   res     1                               ' test brightness
timer                   res     1                               ' bit timer
                                                                 
chpntr                  res     1                               ' hub pointer to levels
chmask                  res     1                               ' channel mask
                                                                 
t1                      res     1                               ' work vars
t2                      res     1                                

                        fit     496
                        

dat

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}